#ifndef __HASH_H__
#define __HASH_H__

#include "log.h"

#define BUF_ERROR               0
#define BUF_EXECUTE             1
#define BUF_GET_WAITFOR         2
#define BUF_GET_DEPENDENT_INFO  3
#define BUF_GET_DEPENDENT_VAL   4
#define BUF_UNDO                5

/* an element on the data list */
typedef struct list_ele_s {
  int bno;
  char *data;
  struct list_ele_s *next;
  struct list_ele_s *prev;
} list_ele_t;

/* the data list */
typedef struct list_s {
  list_ele_t *head;
  list_ele_t *tail;
} list_t;

/* functions for the data list */
list_t *init_list();
void put_block_list(list_t *l, int bno);
void get_block_list(list_t *l, int bno);
void free_list(list_t *l);

/* an element on the queue */
typedef struct q_ele_s {
  char *buffer;
  struct q_ele_s *next;
  struct q_ele_s *prev;
} q_ele_t;

/* the queue */
typedef struct q_s {
  int inode;
  int wait_command;
  q_ele_t *head;
  q_ele_t *tail;
  struct q_s *next;
  struct q_s *prev;
} q_t;

/* a list of queues */
typedef struct {
  q_t *head;
  q_t *tail;
} q_list_t;

/* a hash table of lists */
#define QHASHSIZE  10
#define HASHDEP(x) (x % QHASHSIZE)
typedef struct {
  q_list_t list[QHASHSIZE];
} hash_t;

/* contruction functions */
hash_t *init_hash(void);

/* destruction functions */
void free_hash(hash_t *);

/* functions to maniplate the data structures */
int parse_buffer(char *buffer, int flag);
void add_buffer(hash_t *, char *);
void roll_back(hash_t *);

#endif
