#include "hash.h"

const char LOG_STRING[19][32] = {
  "Free inode",
  "Add directory entry",
  "Remove directory entry",
  "Change directory entry",
  "Set the link count",
  "Add an indirect page",
  "Add a direct page",
  "Allocate block in the block map",
  "Allocate inode in the inode map",
  "Move a directory entry",
  "Set an indirect pointer block",
  "Free blocks in an inode",
  "Free fragments in a block",
  "Change the size of a file",
  "Create a symlink name",
  "Set the attributes on an inode",
  "Make a device node",
  "Initialize a directory",
  "Reallocate a block in a dir"};


int get_block(file_p, lbn)
     struct dinode *file_p;
     int lbn;
{
  int ibn;
  int pbn;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  if(lbn < NDADDR) {
    pbn = file_p->di_db[lbn];
  } else {
    lbn -= NDADDR;
    /* single indirect */
    if(lbn < NINDIR(fs)) {
      bp = getdatablk(file_p->di_ib[0], fs->fs_bsize);
      memcpy(&pbn, &(bp->b_un.b_indir[lbn]), sizeof(int));
      bp->b_flags &= ~B_INUSE;
    } else {
      lbn -= NINDIR(fs);
      /* double indirect */
      if(lbn < NINDIR(fs) * NINDIR(fs)) {
	bp = getdatablk(file_p->di_ib[1], fs->fs_bsize);
	ibn = lbn / NINDIR(fs);
	lbn -= ibn * NINDIR(fs);
	memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	bp->b_flags &= ~B_INUSE;

	bp = getdatablk(ibn, fs->fs_bsize);
	ibn = lbn;
	memcpy(&pbn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	bp->b_flags &= ~B_INUSE;
      } else {
	lbn -= NINDIR(fs) * NINDIR(fs);
	/* triple indirect */
	if(lbn < NINDIR(fs) * NINDIR(fs) * NINDIR(fs)) {
	  bp = getdatablk(file_p->di_ib[2], fs->fs_bsize);
	  ibn = lbn / (NINDIR(fs) * NINDIR(fs));
	  lbn -= ibn * (NINDIR(fs) * NINDIR(fs));
	  memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	  bp->b_flags &= ~B_INUSE;
	  
	  bp = getdatablk(ibn, fs->fs_bsize);
	  ibn = lbn / NINDIR(fs);
	  lbn -= ibn * NINDIR(fs);
	  memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	  bp->b_flags &= ~B_INUSE;

	  bp = getdatablk(ibn, fs->fs_bsize);
	  ibn = lbn;
	  memcpy(&pbn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	  bp->b_flags &= ~B_INUSE;
	} else {
	  /* ERROR! */
	}
      }
    }
  }

  return pbn;
}

void clear_block(file_p, lbn)
     struct dinode *file_p;
     int lbn;
{
  int sibn, dibn;
  struct bufarea *bp = NULL;
  struct fs *fs = &sblock;

  if(lbn < NDADDR) {
    /* direct pointer */
    file_p->di_db[lbn] = 0;
  } else {
    lbn -= NDADDR;
    if(lbn < NINDIR(fs)) {
      /* single indirect */
      bp = getdatablk(file_p->di_ib[0], fs->fs_bsize);
      bp->b_un.b_indir[lbn] = 0;
      if(lbn == 0) {
	file_p->di_ib[0] = 0;
      }
    } else {
      lbn -= NINDIR(fs);
      if(lbn < NINDIR(fs) * NINDIR(fs)) {
	/* double indirect */
	bp = getdatablk(file_p->di_ib[1], fs->fs_bsize);
	if(lbn == 0) {
	  file_p->di_ib[1] = 0;
	}

	sibn = lbn / NINDIR(fs);
	lbn -= sibn * NINDIR(fs);
	dibn = bp->b_un.b_indir[sibn];
	if(lbn == 0) {
	  bp->b_un.b_indir[sibn] = 0;
	}
	bp->b_flags &= ~B_INUSE;

	bp = getdatablk(dibn, fs->fs_bsize);
	bp->b_un.b_indir[lbn] = 0;
      } else {
	lbn -= NINDIR(fs) * NINDIR(fs);
	if(lbn < NINDIR(fs) * NINDIR(fs) * NINDIR(fs)) {
	  /* triple indirect */
	  /* not yet implemented */
	}
      }
    }
  }

  if(bp != NULL) {
    dirty(bp);
    bp->b_flags &= ~B_INUSE;
  }
}

void set_block(file_p, lbn, pbn)
     struct dinode *file_p;
     int lbn;
     int pbn;
{
  int ibn;
  struct bufarea *bp = NULL;
  struct fs *fs = &sblock;

  if(lbn < NDADDR) {
    file_p->di_db[lbn] = pbn;
  } else {
    lbn -= NDADDR;
    /* single indirect */
    if(lbn < NINDIR(fs)) {
      bp = getdatablk(file_p->di_ib[0], fs->fs_bsize);
      memcpy(&(bp->b_un.b_indir[lbn]), &pbn, sizeof(int));
    } else {
      lbn -= NINDIR(fs);
      /* double indirect */
      if(lbn < NINDIR(fs) * NINDIR(fs)) {
	bp = getdatablk(file_p->di_ib[1], fs->fs_bsize);
	ibn = lbn / NINDIR(fs);
	lbn -= ibn * NINDIR(fs);
	memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	bp->b_flags &= ~B_INUSE;

	bp = getdatablk(ibn, fs->fs_bsize);
	ibn = lbn;
	memcpy(&(bp->b_un.b_indir[ibn]), &pbn, sizeof(int));
      } else {
	lbn -= NINDIR(fs) * NINDIR(fs);
	/* triple indirect */
	if(lbn < NINDIR(fs) * NINDIR(fs) * NINDIR(fs)) {
	  bp = getdatablk(file_p->di_ib[2], fs->fs_bsize);
	  ibn = lbn / (NINDIR(fs) * NINDIR(fs));
	  lbn -= ibn * (NINDIR(fs) * NINDIR(fs));
	  memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	  bp->b_flags &= ~B_INUSE;
	  
	  bp = getdatablk(ibn, fs->fs_bsize);
	  ibn = lbn / NINDIR(fs);
	  lbn -= ibn * NINDIR(fs);
	  memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	  bp->b_flags &= ~B_INUSE;
	  
	  bp = getdatablk(ibn, fs->fs_bsize);
	  ibn = lbn;
	  memcpy(&(bp->b_un.b_indir[ibn]), &pbn, sizeof(int));
	} else {
	  /* ERROR! */
	}
      }
    }
  }

  if(bp != NULL) {
    dirty(bp);
    bp->b_flags &= ~B_INUSE;
  }
  inodirty();
}

void free_inode(inum)
     int inum;
{
  int cg;
  struct fs *fs = &sblock;

  /*
   * get the cylinder group
   */
  cg = ino_to_cg(&sblock, inum);
  getblk(&cgblk, cgtod(&sblock, cg), sblock.fs_cgsize);

  /*
   * update the inode map to exclude this inode
   */
  clrbit(cg_inosused(&cgrp), inum % sblock.fs_ipg);

  /* ? */
  cgdirty();
  sbdirty();
}

void alloc_inode(inum)
     int inum;
{
  int cg;
  struct bufarea *bp;
  struct dinode *dp;

  /*
   * get the cylinder group
   */
  cg = ino_to_cg(&sblock, inum);
  getblk(&cgblk, cgtod(&sblock, cg), sblock.fs_cgsize);

  /*
   * update the inode map to include this inode
   */
  setbit(cg_inosused(&cgrp), inum % sblock.fs_ipg);

  /*
   * set the cylinder group and superblock to be dirty
   */
  cgdirty();
  sbdirty();
}

void clear_dir_entry(dir_p, offset, dir_entry_p)
     struct dinode *dir_p;
     int offset;
     struct direct *dir_entry_p;
{
  int lbn;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  /*
   * calculate the logical block number and get the block
   */
  lbn = (offset / fs->fs_bsize);
  bp = getdatablk(get_block(dir_p, lbn), fs->fs_bsize);

  /*
   * take the offset and read off the directory entry
   */
  offset = offset - (lbn * fs->fs_bsize);
  memcpy(dir_entry_p, bp->b_un.b_buf + offset, 8);
  memcpy(dir_entry_p + 8, bp->b_un.b_buf + offset + 8, dir_entry_p->d_namlen);

  /*
   * now simply zero out the area (?)
   */
  memset(bp->b_un.b_buf+offset, 0, 8 + dir_entry_p->d_namlen);

  /*
   * and write back the buffer to disk
   */
  dirty(bp);
  bp->b_flags &= ~B_INUSE;
}

int free_dir_entry(dir_p, offset, count, size)
     struct dinode *dir_p;
     int offset;
     int count;
     int size;
{
  int lbn, inum;
  struct bufarea *bp;
  struct direct *dir_entry_p, *ep;
  struct fs *fs = &sblock;

  /*
   * calculate the logical block number and get the block
   */
  lbn = (offset / fs->fs_bsize);
  bp = getdatablk(get_block(dir_p, lbn), fs->fs_bsize);

  /*
   * take the offset and set the directory entry
   */
  offset = offset - (lbn * fs->fs_bsize);
  dir_entry_p = (struct direct *)(bp->b_un.b_buf + offset);

  /*
   * get the inode number for this entry
   */
  inum = dir_entry_p->d_ino;

  /*
   * now check the count, if 0 set the inode number to 0 and done
   * otherwise we need to colapse the count into the previous entry
   */
  if(count == 0) {
    dir_entry_p->d_ino = 0;
    dir_entry_p->d_reclen = DIRBLKSIZ;
  } else {
    dir_entry_p->d_ino = 0;
    ep = (struct direct *)(bp->b_un.b_buf + offset - count);
    ep->d_reclen = count + size;
  }

  /*
   * zero out the old entry
   */
  memset(dir_entry_p, 0, dir_entry_p->d_reclen);

  /*
   * write the block back to disk
   */
  dirty(bp);
  bp->b_flags &= ~B_INUSE;

  return (inum);
}

int change_dir_entry(dir_p, offset, inum)
     struct dinode *dir_p;
     int offset;
     int inum;
{
  int lbn, retino;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  /*
   * get the block
   */
  lbn = (offset / fs->fs_bsize);
  bp = getdatablk(get_block(dir_p, lbn), fs->fs_bsize);

  /*
   * copy the entry
   */
  offset = offset - lbn * fs->fs_bsize;
  memcpy(&retino, bp->b_un.b_buf + offset, sizeof(int));
  memcpy(bp->b_un.b_buf + offset, &inum, sizeof(int));

  dirty(bp);
  bp->b_flags &= ~B_INUSE;
  
  return (retino);
}

void set_dir_entry(dir_p, offset, dir_entry_pp)
     struct dinode *dir_p;
     int offset;
     struct direct **dir_entry_pp;
{
  int lbn;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  /*
   * get the block
   */
  lbn = (offset / fs->fs_bsize);
  bp = getdatablk(get_block(dir_p, lbn), fs->fs_bsize);
  
  /*
   * copy the entry
   */
  offset = offset - lbn * fs->fs_bsize;
  memcpy(bp->b_un.b_buf + offset, *dir_entry_pp, 8);
  memcpy(bp->b_un.b_buf + offset + 8, *dir_entry_pp + 8, (*dir_entry_pp)->d_namlen);

  /*
   * set the block
   */
  dirty(bp);
  bp->b_flags &= ~B_INUSE;
}

void alloc_frags(blkno, frags)
     int blkno;
     int frags;
{
  int cg, i;
  char *bp;

  if(frags == 0) {
    return;
  }

  /*
   * get the cylinder group (cg)
   */
  cg = dtog(&sblock, blkno);
  getblk(&cgblk, cgtod(&sblock, cg), sblock.fs_cgsize);

  /*
   * simply clear the frag bits, and worry about structure cleanup later
   */
  for(i = 0; i < frags; i++) {
    clrbit(cg_blksfree(&cgrp), dtogd(&sblock, blkno) + i);
  }
  
  cgdirty();
}

void free_frags(blkno, frags)
     int blkno;
     int frags;
{
  int cg, i;

  if(frags == 0) {
    return;
  }

  /*
   * get the cylinder group
   */
  cg = dtog(&sblock, blkno);
  getblk(&cgblk, cgtod(&sblock, cg), sblock.fs_cgsize);

  /*
   * simply set the frag bits, and worry about structure cleanup later
   */
  for(i = 0; i < frags; i++) {
    setbit(cg_blksfree(&cgrp), dtogd(&sblock, blkno) + i);
  }
  
  cgdirty();
}

/*
 * this function retrieves <size> data from inode <log_p> at <offset> and
 * places it into <buffer>
 * it also updates the offset
 */
void get_buffer_data(fs, log_p, buffer, offset, size)
     struct fs *fs;
     struct dinode *log_p;
     char *buffer;
     int *offset;
     int size;
{
  int loff, lbn;
  struct bufarea *bp;

  /*
   * determine the block number and offset in that block
   */
  lbn = (*offset) / fs->fs_bsize;
  loff = (*offset) % fs->fs_bsize;

  /*
   * now grab the appropriate buffer(s) and do the copying
   */
  if(size > fs->fs_bsize - loff) {
    /* going to need 2 buffers to do the copy */
    bp = getdatablk(get_block(log_p, lbn), fs->fs_bsize);
    memcpy(buffer, bp->b_un.b_buf + loff, fs->fs_bsize - loff);
    bp->b_flags &= ~B_INUSE;

    lbn = (lbn + 1) % (log_p->di_size / fs->fs_bsize);
    bp = getdatablk(get_block(log_p, lbn), fs->fs_bsize);
    memcpy(buffer + fs->fs_bsize - loff, bp->b_un.b_buf,
	   size + loff - fs->fs_bsize);
    bp->b_flags &= ~B_INUSE;
  } else {
    /* only 1 buffer will be needed... much easier! */
    bp = getdatablk(get_block(log_p, lbn), fs->fs_bsize);
    memcpy(buffer, bp->b_un.b_buf + loff, size);
    bp->b_flags &= ~B_INUSE;
  }

  /* increase the offset */
  (*offset) = ((*offset) + size) % log_p->di_size;
}

int check_checksum(buffer, size)
     char *buffer;
     int size;
{
  int *buf = (int *)buffer;
  int sum = 0, i;

  for(i = 0; i < size; i++) {
    sum += buf[i];
  }
  if(sum == buf[size])
    return 1;

#ifdef DEBUG_LOGFSCK
  printf("bad checksum (%d; %d): %d %d\n", buf[0], size, sum, buf[size]);
  for(i = 0; i < size; i++) {
    printf("\t%d\n", buf[i]);
  }
#endif

  return 0;  
}


int get_log_entry(fs, log_p, index, buffer)
     struct fs *fs;
     struct dinode *log_p;
     int *index;
     char *buffer;
{
  int command = 0, i, j;
  int cksum = 0;
  int ret = 0;
  struct bufarea *bp;

  /* get the command */
  get_buffer_data(fs, log_p, buffer, index, sizeof(int));
  command = ((int *)(buffer))[0];

  switch(command) {
  case(LOGGING_ADDDIRECT):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*6);
    cksum = 6;
    ret = sizeof(int)*7;
    break;
  case(LOGGING_UPDATEBLKMAP):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*6);
    cksum = 6;
    ret = sizeof(int)*7;
    break;
  case(LOGGING_UPDATEINOMAP):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*5);
    cksum = 5;
    ret = sizeof(int)*6;
    break;
  case(LOGGING_FREEINODE):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*4);
    cksum = 4;
    ret = sizeof(int)*5;
    break;
  case(LOGGING_ADDDIRENTRY):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*8 + 8);
    cksum = 8;

    i = (buffer + sizeof(int)*9)[7];
    get_buffer_data(fs, log_p, buffer+sizeof(int)*9 + 8, index, i);

    ret = sizeof(int)*9 + 8 + i;
    break;
  case(LOGGING_REMDIRENTRY):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*10);
    cksum = 10;
    ret = sizeof(int)*11;
    break;
  case(LOGGING_CHANGEDIRENTRY):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*10);
    cksum = 10;
    ret = sizeof(int)*11;
    break;
  case(LOGGING_INCLINKCNT):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*5);
    cksum = 5;
    ret = sizeof(int)*6;
    break;
  case(LOGGING_SETINDIRMETA):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*7);
    cksum = 7;
    ret = sizeof(int)*8;
    break;
  case(LOGGING_ADDINDIRPAGE):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*7);
    cksum = 7;
    ret = sizeof(int)*8;
    break;
  case(LOGGING_CHANGESIZE):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*6);
    cksum = 6;
    ret = sizeof(int)*7;
    break;
  case(LOGGING_FREEFRAGS):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*6);
    cksum = 6;
    ret = sizeof(int)*7;
    break;
  case(LOGGING_FREEBLOCKS):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*6);
    cksum = 6;
    ret = sizeof(int)*7;
    break;
  case(LOGGING_SYMLINK):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*5);
    cksum = 5;

    i = ((int *)(buffer))[4];
    get_buffer_data(fs, log_p, buffer+sizeof(int)*6, index, i);

    ret = sizeof(int)*6 + i;
    break;
  case(LOGGING_MOVDIRENTRY):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*8 + 8);
    cksum = 8;

    i = (buffer + sizeof(int)*9)[7];
    get_buffer_data(fs, log_p, buffer+sizeof(int)*9 + 8, index, i);

    ret = sizeof(int)*9 + 8 + i;
    break;
  case(LOGGING_SETATTR):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*8);
    cksum = 8;
    ret = sizeof(int)*9;
    break;
  case(LOGGING_MKNOD):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*5);
    cksum = 5;
    ret = sizeof(int)*6;
    break;
  case(LOGGING_INITDIR):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*7);
    cksum = 7;
    ret = sizeof(int)*8;
    break;
  case(LOGGING_REALLOCBUF):
    get_buffer_data(fs, log_p, buffer+sizeof(int), index, sizeof(int)*8);
    cksum = 8;
    
    i = ((int *)(buffer))[6];
    get_buffer_data(fs, log_p, buffer+sizeof(int)*9, index, i);
    
    ret = sizeof(int)*9 + i;
    break;
  default:
    ret = 0;
#ifdef DEBUG_LOGFSCK
    printf("command was: %d\n", command);
    printf("break 1\n");
#endif
    break;
  }

  if(cksum) {
    cksum = check_checksum(buffer, cksum);
    if(!cksum) {
      ret = 0;

#ifdef DEBUG_LOGFSCK
      printf("break 2\n");
#endif
    }
  }

  return ret;
}

/* global variable */
list_t *l;

void log_roll_forward() {
  struct dinode *log_p, *file_p, *dir_p, log_inode;
  struct bufarea *bp, *dir_bp;
  unsigned int log_size, log_start, log_end;
  unsigned int first_block, last_block;
  unsigned int i, j, k;
  unsigned int num_log_blocks;
  unsigned int inum;
  int old_sec, old_nsec;
  int command, *log_buf;
  int extra_frag;
  struct fs *fs = &sblock;
  int lbn = 0, offset = 0, dlbn = 0;
  struct direct dir_entry, *dir_entry_p;
  char log_buffer[sizeof(int)*11 + 8 + 256 + 8192];
  char *tbuf;
  int size;

  /*
   * create the hash table and the data list
   */
  hash_t *h = init_hash();
  l = init_list();

  /*
   * for now, just roll through each entry and print it out
   */
  offset = fs->fs_log_start;

#ifdef DEBUG_LOGFSCK
  printf("start of the log: %d\n", offset);
#endif

  /* get the inode */
  log_p = ginode(LOGGING_INO);
  memcpy(&log_inode, log_p, sizeof(struct dinode));
  old_sec = 0; old_nsec = 0;
  while(size = get_log_entry(fs, &log_inode, &offset, &log_buffer)) {
    if(old_sec > ((int *)(log_buffer))[1] ||
       (old_sec == ((int *)(log_buffer))[1] &&
	old_nsec > ((int *)(log_buffer))[2])) {
#ifdef DEBUG_LOGFSCK
      printf("old time: %d:%d\n", old_sec, old_nsec);
      printf("entry time (%d): %d:%d\n",
	     ((int *)(log_buffer))[0],
	     ((int *)(log_buffer))[1],
	     ((int *)(log_buffer))[2]);
      printf("break 3\n");
#endif
      break;
    }

    old_sec = ((int *)(log_buffer))[1];
    old_nsec = ((int *)(log_buffer))[2];

    /*
     * we have the entry, lets put it in the hash table
     * and let it do all the work
     */
    tbuf = (char *)malloc(size);
    memcpy(tbuf, log_buffer, size);
    add_buffer(h, tbuf);
  }

#ifdef DEBUG_LOGFSCK
  printf("End of log before or at: %d\n", offset);
#endif

  /*
   * roll back any uncompleted operations
   */
  roll_back(h);

  /*
   * free the hash table and the data list
   */
  free(h);
  free_list(l);
}
