#include "log.h"

int get_block(file_p, lbn)
     struct dinode *file_p;
     int lbn;
{
  int ibn;
  int pbn;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  if(lbn < NDADDR) {
    pbn = file_p->di_db[lbn];
    printf("\t\tdirect: %d\n", pbn);
  } else {
    lbn -= NDADDR;
    /* single indirect */
    if(lbn < NINDIR(fs)) {
      printf("\t\tsingle indirect: %d\n", file_p->di_ib[0]);
      bp = getdatablk(file_p->di_ib[0], fs->fs_bsize);
      printf("\t\tdirect: %d\n", bp->b_un.b_indir[lbn]);
      memcpy(&pbn, &(bp->b_un.b_indir[lbn]), sizeof(int));
    } else {
      lbn -= NINDIR(fs);
      /* double indirect */
      if(lbn < NINDIR(fs) * NINDIR(fs)) {
	bp = getdatablk(file_p->di_ib[1], fs->fs_bsize);
	ibn = lbn / NINDIR(fs);
	lbn -= ibn * NINDIR(fs);
	memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));

	bp = getdatablk(ibn, fs->fs_bsize);
	ibn = lbn;
	memcpy(&pbn, &(bp->b_un.b_indir[ibn]), sizeof(int));
      } else {
	lbn -= NINDIR(fs) * NINDIR(fs);
	/* triple indirect */
	if(lbn < NINDIR(fs) * NINDIR(fs) * NINDIR(fs)) {
	  bp = getdatablk(file_p->di_ib[2], fs->fs_bsize);
	  ibn = lbn / (NINDIR(fs) * NINDIR(fs));
	  lbn -= ibn * (NINDIR(fs) * NINDIR(fs));
	  memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	  
	  bp = getdatablk(ibn, fs->fs_bsize);
	  ibn = lbn / NINDIR(fs);
	  lbn -= ibn * NINDIR(fs);
	  memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	  
	  bp = getdatablk(ibn, fs->fs_bsize);
	  ibn = lbn;
	  memcpy(&pbn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	} else {
	  /* ERROR! */
	}
      }
    }
  }

  return pbn;
}

void set_block(file_p, lbn, pbn)
     struct dinode *file_p;
     int lbn;
     int pbn;
{
  int ibn;
  struct bufarea *bp = NULL;
  struct fs *fs = &sblock;

  printf("\t\tset_block: lbn = %d\n", lbn);
  if(lbn < NDADDR) {
    file_p->di_db[lbn] = pbn;
  } else {
    lbn -= NDADDR;
    printf("\t\tset_block: slbn = %d\n", lbn);
    /* single indirect */
    if(lbn < NINDIR(fs)) {
      bp = getdatablk(file_p->di_ib[0], fs->fs_bsize);
      memcpy(&(bp->b_un.b_indir[lbn]), &pbn, sizeof(int));
    } else {
      lbn -= NINDIR(fs);
      printf("\t\tset_block: dlbn = %d\n", lbn);
      /* double indirect */
      if(lbn < NINDIR(fs) * NINDIR(fs)) {
	bp = getdatablk(file_p->di_ib[1], fs->fs_bsize);
	ibn = lbn / NINDIR(fs);
	lbn -= ibn * NINDIR(fs);
	memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));

	bp = getdatablk(ibn, fs->fs_bsize);
	ibn = lbn;
	memcpy(&(bp->b_un.b_indir[ibn]), &pbn, sizeof(int));
      } else {
	lbn -= NINDIR(fs) * NINDIR(fs);
	/* triple indirect */
	if(lbn < NINDIR(fs) * NINDIR(fs) * NINDIR(fs)) {
	  bp = getdatablk(file_p->di_ib[2], fs->fs_bsize);
	  ibn = lbn / (NINDIR(fs) * NINDIR(fs));
	  lbn -= ibn * (NINDIR(fs) * NINDIR(fs));
	  memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	  
	  bp = getdatablk(ibn, fs->fs_bsize);
	  ibn = lbn / NINDIR(fs);
	  lbn -= ibn * NINDIR(fs);
	  memcpy(&ibn, &(bp->b_un.b_indir[ibn]), sizeof(int));
	  
	  bp = getdatablk(ibn, fs->fs_bsize);
	  ibn = lbn;
	  memcpy(&(bp->b_un.b_indir[ibn]), &pbn, sizeof(int));
	} else {
	  /* ERROR! */
	}
      }
    }
  }

  if(bp != NULL) {
    dirty(bp);
    bp->b_flags &= ~B_INUSE;
  }
  inodirty();
}

void free_inode(inum)
     int inum;
{
  int cg;
  struct fs *fs = &sblock;

  /*
   * get the cylinder group
   */
  cg = ino_to_cg(&sblock, inum);
  getblk(&cgblk, cgtod(&sblock, cg), sblock.fs_cgsize);

  /*
   * update the inode map to exclude this inode
   */
  clrbit(cg_inosused(&cgrp), inum % sblock.fs_ipg);

  /* ? */
  cgdirty();
  sbdirty();
}

void alloc_inode(inum)
     int inum;
{
  int cg;
  struct bufarea *bp;
  struct dinode *dp;

  /*
   * get the cylinder group
   */
  cg = ino_to_cg(&sblock, inum);
  getblk(&cgblk, cgtod(&sblock, cg), sblock.fs_cgsize);

  /*
   * update the inode map to include this inode
   */
  setbit(cg_inosused(&cgrp), inum % sblock.fs_ipg);

  /*
   * set the cylinder group and superblock to be dirty
   */
  cgdirty();
  sbdirty();
}

void clear_dir_entry(dir_p, offset, dir_entry_p)
     struct dinode *dir_p;
     int offset;
     struct direct *dir_entry_p;
{
  int lbn;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  /*
   * calculate the logical block number and get the block
   */
  lbn = (offset / fs->fs_bsize);
  bp = getdatablk(get_block(dir_p, lbn), fs->fs_bsize);

  /*
   * take the offset and read off the directory entry
   */
  printf("offset: %d\n", offset);
  printf("lbn: %d\n", lbn);
  offset = offset - (lbn * fs->fs_bsize);
  printf("offset: %d\n", offset);
  memcpy(dir_entry_p, bp->b_un.b_buf + offset, 8);
  memcpy(dir_entry_p + 8, bp->b_un.b_buf + offset + 8, dir_entry_p->d_namlen);

  /*
   * now simply zero out the area (?)
   */
  memset(bp->b_un.b_buf+offset, 0, 8 + dir_entry_p->d_namlen);

  /*
   * and write back the buffer to disk
   */
  dirty(bp);
  bp->b_flags &= ~B_INUSE;
}

int free_dir_entry(dir_p, offset, count, size)
     struct dinode *dir_p;
     int offset;
     int count;
     int size;
{
  int lbn, inum;
  struct bufarea *bp;
  struct direct *dir_entry_p, *ep;
  struct fs *fs = &sblock;

  /*
   * calculate the logical block number and get the block
   */
  lbn = (offset / fs->fs_bsize);
  printf("\t\tlogical block number to be freed: %d\n", lbn);
  printf("\t\tpbn: %d\n", get_block(dir_p, lbn));
  bp = getdatablk(get_block(dir_p, lbn), fs->fs_bsize);

  /*
   * take the offset and set the directory entry
   */
  offset = offset - (lbn * fs->fs_bsize);
  dir_entry_p = (struct direct *)(bp->b_un.b_buf + offset);

  printf("\t\tgot a directory entry: %d, %d, %d, %d\n", dir_entry_p->d_ino,
	 dir_entry_p->d_reclen, dir_entry_p->d_type, dir_entry_p->d_namlen);

  /*
   * get the inode number for this entry
   */
  inum = dir_entry_p->d_ino;

  /*
   * now check the count, if 0 set the inode number to 0 and done
   * otherwise we need to colapse the count into the previous entry
   */
  printf("\t\tthe count: %d\n", count);
  if(count == 0) {
    dir_entry_p->d_ino = 0;
  } else {
    dir_entry_p->d_ino = 0;
    ep = (struct direct *)(bp->b_un.b_buf + offset - count);
    printf("previous entry inode: %d\n", ep->d_ino);
    ep->d_reclen = count + size;
    printf("new reclen: %d\n", ep->d_reclen);
  }

  /*
   * zero out the old entry
   */
  memset(dir_entry_p, 0, dir_entry_p->d_reclen);

  /*
   * write the block back to disk
   */
  dirty(bp);
  bp->b_flags &= ~B_INUSE;

  return (inum);
}

int change_dir_entry(dir_p, offset, inum)
     struct dinode *dir_p;
     int offset;
     int inum;
{
  int lbn, retino;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  /*
   * get the block
   */
  lbn = (offset / fs->fs_bsize);
  printf("\t\tlbn = %d\n", lbn);
  bp = getdatablk(get_block(dir_p, lbn), fs->fs_bsize);

  /*
   * copy the entry
   */
  offset = offset - lbn * fs->fs_bsize;
  memcpy(&retino, bp->b_un.b_buf + offset, sizeof(int));
  memcpy(bp->b_un.b_buf + offset, &inum, sizeof(int));

  dirty(bp);
  bp->b_flags &= ~B_INUSE;
  
  return (retino);
}

void set_dir_entry(dir_p, offset, dir_entry_pp)
     struct dinode *dir_p;
     int offset;
     struct direct **dir_entry_pp;
{
  int lbn;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  /*
   * get the block
   */
  lbn = (offset / fs->fs_bsize);
  bp = getdatablk(get_block(dir_p, lbn), fs->fs_bsize);
  
  /*
   * copy the entry
   */
  offset = offset - lbn * fs->fs_bsize;
  memcpy(bp->b_un.b_buf + offset, *dir_entry_pp, 8);
  memcpy(bp->b_un.b_buf + offset + 8, *dir_entry_pp + 8, (*dir_entry_pp)->d_namlen);

  /*
   * set the block
   */
  dirty(bp);
  bp->b_flags &= ~B_INUSE;
}

void alloc_frags(blkno, frags)
     int blkno;
     int frags;
{
  int cg, i;
  char *bp;

  /*
   * get the cylinder group (cg)
   */
  cg = dtog(&sblock, blkno);
  getblk(&cgblk, cgtod(&sblock, cg), sblock.fs_cgsize);

  /*
   * simply clear the frag bits, and worry about structure cleanup later
   */
  for(i = 0; i < frags; i++) {
    clrbit(cg_blksfree(&cgrp), dtogd(&sblock, blkno) + i);
  }
  
  cgdirty();
}

void free_frags(blkno, frags)
     int blkno;
     int frags;
{
  int cg, i;

  if(frags == 0) {
    return;
  }

  /*
   * get the cylinder group
   */
  cg = dtog(&sblock, blkno);
  getblk(&cgblk, cgtod(&sblock, cg), sblock.fs_cgsize);

  /*
   * simply set the frag bits, and worry about structure cleanup later
   */
  for(i = 0; i < frags; i++) {
    setbit(cg_blksfree(&cgrp), dtogd(&sblock, blkno) + i);
  }
  
  cgdirty();
}

void get_buffer_data(fs, log_p, buffer, bp, lbn, offset, size)
     struct fs *fs;
     struct dinode *log_p;
     char *buffer;
     struct bufarea **bp;
     int *lbn;
     int offset;
     int size;
{
  int add = 0;
  /*
   * figure out if the size will cross block boundaries
   */
  if(offset + size <= fs->fs_bsize) {
    /*
     * if it doesn't just copy out
     */
    /*    printf("\t\t\t*** reading buffer for lbn %d at offset %d\n", *lbn, offset); */
    memcpy(buffer, (*bp)->b_un.b_buf+offset, size);

    printf("final size: %d %d\n", offset + size, fs->fs_bsize);
    if((offset + size) == fs->fs_bsize) {
      printf("YAYA\n");
      (*lbn)++;
      (*bp) = getdatablk(get_block(log_p, *lbn), fs->fs_bsize);
    }
  } else {
    int p1 = fs->fs_bsize - offset;
    int p2 = size - p1;

    /*
     * if it does, we need to get the next block too and copy some from there
     */
    printf("offset: %d, size %d\n", offset, size);
    memcpy(buffer, (*bp)->b_un.b_buf+offset, p1);
    printf("buffer-> 0: %d, 1: %d, 2: %d\n", 
	   ((int *)buffer)[0], ((int *)buffer)[1], ((int *)buffer)[2]);
    /*    printf("\t\t\t*** reading buffer for lbn %d and ", *lbn); */
    (*lbn)++;
    if((*lbn + 1) * fs->fs_bsize > log_p->di_size) {
      *lbn = 0;
      add = LOGGING_STARTPOINT;
    } else {
      add = 0;
    }

    printf("lbn: %d, add: %d\n, p1: %d, p2: %d", *lbn, add, p1, p2);
    *bp = getdatablk(get_block(log_p, *lbn), fs->fs_bsize);

    memcpy(buffer + p1, (*bp)->b_un.b_buf, p2);
    printf("buffer-> 0: %d, 1: %d, 2: %d\n", 
	   ((int *)buffer)[0], ((int *)buffer)[1], ((int *)buffer)[2]);
    printf("buffer-> %d", buffer[p1]);
  }
}
  
void log_roll_forward() {
  struct dinode *log_p, *file_p, *dir_p;
  struct bufarea *bp, *dir_bp;
  unsigned int log_size, log_start, log_end;
  unsigned int first_block, last_block;
  unsigned int i, j, k;
  unsigned int num_log_blocks;
  unsigned int inum;
  int command, log_buf[6];
  int extra_frag;
  struct fs *fs = &sblock;
  int lbn = 0, offset = 0, dlbn = 0;
  struct direct dir_entry, *dir_entry_p;

  /*
   * get the log inode
   */
  log_p = ginode((ino_t)LOGGING_INO);
  
  /*
   * read in the first block of the log
   */
  i = 0;
  printf("CRUCIAL -- READING IN THE LOG FILE BLOCK %d : %d; %d; %d\n", i,
	 log_p->di_size, fs->fs_bsize, (log_p->di_size / fs->fs_bsize));
  bp = getdatablk(get_block(log_p, lbn), fs->fs_bsize);
  
  /*
   * grab the start and size of the log on this block
   */ 
  log_start = 0; /* fs->fs_log_start; */
  log_end = fs->fs_log_end;

  /* set the logical block number */
  lbn = log_start / fs->fs_bsize;
  printf("lbn: %d\n", lbn);
  
  printf("log start: %d\n", log_start);
  printf("log end: %d\n", log_end);
  
  /*
   * loop through the log entries in the block
   */
  j = log_start;
  while(j != log_end) {
    /*
     * get the first command
     */
    get_buffer_data(fs, log_p, &command, &bp, &lbn,
		    j % fs->fs_bsize, sizeof(int));
    j += sizeof(int);

    printf("Got a command: %d %d\n", command, lbn);
    
    switch(command) {
    case(LOGGING_ADDDIRECT):
      printf("Adding direct pointer\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*4);
      j += sizeof(int)*4;
      
      /*
       * get the proper dinode
       */
      printf("\tgetting the inode %d\n", log_buf[0]);
      file_p = ginode(log_buf[0]);
      
      /*
       * correct the file's size?
       */
      /*
      printf("\tsetting the file's size to be %d\n", log_buf[1] * fs->fs_bsize + log_buf[3]);
      file_p->di_size = log_buf[1] * fs->fs_bsize + log_buf[3]; 
      extra_frag = (file_p->di_size % fs->fs_fsize) ? 1 : 0; 
      file_p->di_blocks = ((file_p->di_size / fs->fs_fsize) + extra_frag) * btodb(fs->fs_fsize);
      */
      /*
       * set the block to the proper lbn
       * direct pointers only deal with things in the inode
       * no indirect page checking...
       *//*
      printf("\tfile_p->di_blocks = %d\n", file_p->di_blocks);
	 */
      printf("\tthe inode: %d, the lbn: %d, the pbn: %d\n", log_buf[0], log_buf[1], log_buf[2]);

      /*
	set_block(file_p, log_buf[1], log_buf[2]);
      */
      file_p->di_db[log_buf[1]] = log_buf[2];
      inodirty();
      break;
    case(LOGGING_UPDATEBLKMAP):
      printf("Updating the block map\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*2);
      j += sizeof(int)*2;
	
      /*
       * updating the block map
       */
      printf("trying to allocate fragments starting at %d and going for %d\n", log_buf[0], log_buf[1]);
      alloc_frags(log_buf[0], log_buf[1]);
      break;
    case(LOGGING_UPDATEINOMAP):
      printf("Updating the inode map\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*2);
      j += sizeof(int)*2;

      /*
       * update the inode map
       */
      printf("\tallocating an inode: %d\n", log_buf[0]);
      alloc_inode(log_buf[0]);

      /* "allocate" the inode (zero it out?) */
      printf("\tgetting an inode: %d\n", log_buf[0]);
      file_p = ginode(log_buf[0]);
      printf("\tgot an inode: %d\n", file_p->di_inumber);
      file_p->di_flags = 0;
      printf("\tsetting the mode to be: %d\n", log_buf[1]);
      file_p->di_mode = log_buf[1];
      file_p->di_atime = time(NULL);
      file_p->di_mtime = file_p->di_ctime = file_p->di_atime;
      file_p->di_size = 0;
      file_p->di_blocks = 0;
      inodirty();
	
      break;
    case(LOGGING_FREEINODE):
      printf("Freeing an inode\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int));
      j += sizeof(int);

      /*
       * set the info for the inode to be zero'ed out
       */
      printf("\tInode: %d\n", log_buf[0]);
      file_p = ginode(log_buf[0]);
      file_p->di_nlink = 0;
      file_p->di_mode = 0;
      file_p->di_size = 0;
      file_p->di_blocks = 0;
      file_p->di_flags = 0;
      inodirty();

      /*
       * update the inode map
       */
      free_inode(log_buf[0]);

      break;
    case(LOGGING_ADDDIRENTRY):
      printf("Adding a directory entry\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*3);
      j += sizeof(int)*3;

      /*
       * get the directory entry
       */
      get_buffer_data(fs, log_p, &dir_entry, &bp, &lbn,
		      j % fs->fs_bsize, 8);
      get_buffer_data(fs, log_p, &(dir_entry.d_name), &bp, &lbn,
		      (j+8) % fs->fs_bsize, dir_entry.d_namlen);
      dir_entry.d_name[dir_entry.d_namlen] = 0;

      printf("\tInode: %d\n", log_buf[0]);
      printf("\tat offset: %d\n", log_buf[1]);
      printf("\tentry ->\n");
      printf("\t\tinode %d\n", dir_entry.d_ino);
      printf("\t\trecord length %d\n", dir_entry.d_reclen);
      printf("\t\ttype %d\n", dir_entry.d_type);
      printf("\t\tname length %d\n", dir_entry.d_namlen);
      printf("\t\tname %s\n", dir_entry.d_name);

      /*
       * increase the index to match the size of the directory entry
       */
      j += 8 + dir_entry.d_namlen;

      /* 
       * get the dinode and the lbn
       */
      offset = log_buf[1];
      printf("\tgetting an inode %d\n", log_buf[0]);
      dir_p = ginode(log_buf[0]);
      dlbn = offset/fs->fs_bsize;

      /*
       * get the block from the offset
       */
      printf("\tlogical block number %d\n", dlbn);
      dir_bp = getdatablk(get_block(dir_p, dlbn), fs->fs_bsize);

      /*
       * update the directory entry
       */
      offset = offset - dlbn * fs->fs_bsize;
      memcpy(dir_bp->b_un.b_buf +offset, &dir_entry, 8 + dir_entry.d_namlen);

      /*
       * need to correct the previous entry's rec_len
       */
      if(log_buf[2] != 0) {
	short temp;
	/*
	 * we have the previous entry's size, set it
	 */
	printf("previous entry's size: %d\n", log_buf[2]);
	temp = log_buf[2];
	memcpy(dir_bp->b_un.b_buf + offset - log_buf[2] + 4,
	       &(temp), sizeof(short));
      } else {
	/*
	 * we are the first entry in the directory
	 * we should update the size?
	 */
	/* note: 512 temporary */
	dir_p->di_size = ((offset / 512) + 1) * 512;
	printf("directory inode size: %d\n", dir_p->di_size);
	inodirty();
      }

      /* 
       * write out the block
       */
      dirty(dir_bp);
      break;
    case(LOGGING_MOVDIRENTRY):
      printf("Moving a directory entry\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*5);
      j += sizeof(int)*5;

      /*
       * get the dinode of the directory
       */
      printf("\tgetting an inode %d\n", log_buf[0]);
      dir_p = ginode(log_buf[0]);

      /*
       * get the old entry
       */
      printf("log_buf[1]: %d\n", log_buf[1]);
      printf("log_buf[2]: %d\n", log_buf[2]);
      printf("log_buf[3]: %d\n", log_buf[3]);
      printf("log_buf[4]: %d\n", log_buf[4]);

      clear_dir_entry(dir_p, log_buf[2], &dir_entry);

      /*
       * set the new entry
       */
      set_dir_entry(dir_p, log_buf[3], &dir_entry);
      break;
    case(LOGGING_REMDIRENTRY):
      printf("Removing a directory entry\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*4);
      j += sizeof(int)*4;

      /*
       * get the dinode of the directory
       */
      printf("\tgetting an inode %d\n", log_buf[0]);
      dir_p = ginode(log_buf[0]);

      printf("\tabout to remove entry at offset %d\n", log_buf[1]);
      printf("\tabout to update entry at offset %d\n", log_buf[1] - log_buf[2]);
      printf("\tinode reclen: %d\n", log_buf[3]);

      /*
       * clear the directory entry at the given offset
       */
      inum = free_dir_entry(dir_p, log_buf[1], log_buf[2], log_buf[3]);
      break;
    case(LOGGING_CHANGEDIRENTRY):
      printf("Changing a directory entry\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*4);
      j += sizeof(int)*4;

      /*
       * get the dinode of the directory
       */
      printf("\tgetting an inode %d\n", log_buf[0]);
      dir_p = ginode(log_buf[0]);

      /*
       * change the directory entry
       */
      inum = change_dir_entry(dir_p, log_buf[1], log_buf[2]);

      /*
       * updating the inode to the proper link count
       */
      printf("\tgetting an inode %d and setting link to %d\n", inum, log_buf[3]);

      /*
	file_p = ginode(inum);
	file_p->di_nlink = log_buf[3];
	inodirty();
      */
      break;
    case(LOGGING_INCLINKCNT):
      printf("Increasing the link count on a file\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*2);
      j += sizeof(int)*2;
	
      /* increase the link count on the given inode */
      printf("\tgetting inode %d\n", log_buf[0]);
      file_p = ginode(log_buf[0]);
      printf("\tgot inode %d\n", file_p->di_inumber);
      file_p->di_nlink = log_buf[1];
      printf("\tset the link count to %d\n", file_p->di_nlink);
      inodirty();
      break;
    case(LOGGING_SETINDIRMETA):
      printf("Setting the indirect meta data\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*3);
      j += sizeof(int)*3;
      break;
    case(LOGGING_ADDINDIRPAGE):
      printf("Adding an indirect meta data page\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*4);
      j += sizeof(int)*4;

      /*
       * get the dinode of the file
       */
      printf("\tgetting an inode %d\n", log_buf[0]);
      file_p = ginode(log_buf[0]);

      /*
       * update the proper size and # of blocks
       */
      file_p->di_size = log_buf[1] * fs->fs_bsize + log_buf[3];
      extra_frag = (file_p->di_size % fs->fs_fsize) ? 1 : 0;
      printf("\tEXTRA FRAG: %d\n", extra_frag);
      file_p->di_blocks = ((file_p->di_size / fs->fs_fsize) + extra_frag) * btodb(fs->fs_fsize);
      if(log_buf[1] > NDADDR) {
	/* with single indirect, we always have a full block? */
	extra_frag = (file_p->di_size % fs->fs_bsize) ? 1 : 0;
	file_p->di_blocks = ((file_p->di_size / fs->fs_bsize) + extra_frag) * fs->fs_frag * btodb(fs->fs_fsize);

	/* single indirect */
	file_p->di_blocks += fs->fs_frag * btodb(fs->fs_fsize);
      }
      if(log_buf[1] > NINDIR(fs) + NDADDR) {
	/* double indirect */
	file_p->di_blocks += fs->fs_frag * btodb(fs->fs_fsize);
	  
	/* single indirect's within the double indirect */
	file_p->di_blocks += (((log_buf[1] - NINDIR(fs) - NDADDR) / NINDIR(fs)) + 1)  * fs->fs_frag * btodb(fs->fs_fsize);
      }
      if(log_buf[1] > NINDIR(fs)*NINDIR(fs) + NDADDR) {
	/* triple indirect */
	file_p->di_blocks += fs->fs_frag * btodb(fs->fs_fsize);

	/* double's within triple */
	file_p->di_blocks += (((log_buf[1] - NINDIR(fs)*NINDIR(fs) - NINDIR(fs) - NDADDR) / (NINDIR(fs)*NINDIR(fs))) + 1)  * fs->fs_frag * btodb(fs->fs_fsize);
	  
	/* single's within doubles */
      }



      printf("\tthe size %d; the number of blocks %d, %d, %d, %d\n", 
	     file_p->di_size, file_p->di_blocks, fs->fs_fsize, btodb(fs->fs_fsize));

      /*
       * set the logical block number
       */
      set_block(file_p, log_buf[1], log_buf[2]);
      break;
    case(LOGGING_FREEBLOCKS):
      printf("Freeing blocks\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*3);
      j += sizeof(int)*3;

      /* 
       * get the blocks from the inode and 0 out inode entries
       */
      first_block = (log_buf[2] == 0) ? 0 : (log_buf[2] / fs->fs_bsize) + 1;
      last_block = (log_buf[1] / fs->fs_bsize);

      printf("\tfirst_block = %d; last_block = %d\n", first_block, last_block);

      /*
       * get the dinode of the file
       */
      printf("\tgetting an inode %d\n", log_buf[0]);
      file_p = ginode(log_buf[0]);

      /*
       * free all the full blocks
       */
      for(k = first_block; k < last_block; k++) {
	/*
	 * set the block to be free in the block map
	 */
	/*	  free_frags(get_block(file_p, k), 8); */

	/*
	 * clear the inode block pointer
	 */
	set_block(file_p, k, 0);
      }

      /*
       * now free any left over fragments
       */
      printf("\tk = %d\n", k);
      printf("\tblock %d\n", get_block(file_p, k));
      printf("\tlogbuf2 %d\n", log_buf[2]);
      if(get_block(file_p, k) != 0) {
	/*	  free_frags(get_block(file_p, k),
		  ((log_buf[2] % fs->fs_bsize)/fs->fs_fsize)+1); */
      }
      set_block(file_p, k, 0);

      break;
    case(LOGGING_FREEFRAGS):
      printf("Freeing fragments\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*2);
      j += sizeof(int)*2;

      printf("\tfreeing %d fragments: start block %d\n",
	     log_buf[1], log_buf[0]);
      free_frags(log_buf[0], log_buf[1]);
      break;
    case(LOGGING_CHANGESIZE):
      printf("Changing the size of a file\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*3);
      j += sizeof(int)*3;

      printf("\tgetting an inode %d\n", log_buf[0]);
      printf("\tchanged to size %d\n", log_buf[1]);
      printf("\tset num blocks to %d\n", log_buf[2]);
      file_p = ginode(log_buf[0]);
      file_p->di_size = log_buf[1];
      file_p->di_blocks = log_buf[2];
      inodirty();
      break;
    case(LOGGING_SYMLINK):
      printf("Creating a symlink with a short name\n");
      get_buffer_data(fs, log_p, &log_buf, &bp, &lbn,
		      j % fs->fs_bsize, sizeof(int)*2);
      j += sizeof(int)*2;

      printf("\tInode: %d\n\tSize: %d\n", log_buf[0], log_buf[1]);
      file_p = ginode(log_buf[0]);
      file_p->di_size = log_buf[1];
      get_buffer_data(fs, log_p, file_p->di_db, &bp, &lbn,
		      j % fs->fs_bsize, log_buf[1]);
      printf("%s\n", file_p->di_db);
      j += log_buf[1];
      inodirty();
      break;
    case(LOGGING_ENDLOG):
      printf("ENDLOG\n");
      j = 0;
      break;
    default:
      break;
    }
    printf("J = %d\n", j);
  }
  

  /*
   * Here we want to update the counts on the cylinder groups, etc.
   */
  
  printf("Updating counts on the cylinder groups\n");
}
