#include "hash.h"

extern const char LOG_STRING[17][32];
extern list_t *l;

/*
 * this function does all of the buffer parsing and passes back information
 * when requested based upon the passed in flag
 */
int parse_buffer(char *buffer, int flag) {
  int *log_buf = ((int *)(buffer));
  int command = log_buf[0];
  int return_val = -1;

  /* temp */
  struct dinode *log_p, *file_p, *dir_p, log_inode;
  struct bufarea *bp, *dir_bp;
  unsigned int log_size, log_start, log_end;
  unsigned int first_block, last_block;
  unsigned int i, j, k;
  unsigned int num_log_blocks;
  unsigned int inum;
  int old_sec, old_nsec;
  int extra_frag;
  struct fs *fs = &sblock;
  int lbn = 0, offset = 0, dlbn = 0;
  struct direct dir_entry, *dir_entry_p;
  struct dirtemplate dtemp;

#ifdef DEBUG_LOGFSCK
  if(flag == BUF_EXECUTE) {
    printf("executing: (%d) %s\n", log_buf[0], LOG_STRING[log_buf[0]-1]);
  } else if(flag == BUF_UNDO) {
    printf("undoing: (%d) %s\n", log_buf[0], LOG_STRING[log_buf[0]-1]);
  }
#endif

  switch(command) {
  case(LOGGING_FREEINODE):
    if(flag == BUF_EXECUTE) {
      /*
       * set the info for the inode to be zero'ed out
       */
      file_p = ginode(log_buf[3]);
      memset(file_p, 0, sizeof(struct dinode));
      inodirty();

      /*
       * update the inode map to free the inode
       */
      free_inode(log_buf[3]);
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * don't have to undo this operation, since it is a finishing
       * operation, rather than a starting or intermediary
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_ADDDIRENTRY):
    if(flag == BUF_EXECUTE) {
      /*
       * get the directory entry
       */
      memcpy(&dir_entry, buffer + sizeof(int)*9, 8);
      memcpy(&(dir_entry.d_name), buffer + sizeof(int)*9 + 8,
	     dir_entry.d_namlen);
      dir_entry.d_name[dir_entry.d_namlen] = 0;

      /* 
       * get the dinode, set the link count and the lbn
       */
      dir_p = ginode(log_buf[3]);
      dir_p->di_nlink = log_buf[4];
      dlbn = log_buf[6]/fs->fs_bsize;
      offset = log_buf[6] - dlbn * fs->fs_bsize;

      /*
       * update the directory entry
       */
      dir_bp = getdatablk(get_block(dir_p, dlbn), fs->fs_bsize);
      memcpy(dir_bp->b_un.b_buf + offset, &dir_entry, 8+dir_entry.d_namlen+1);

      /*
       * need to correct the previous entry's rec_len
       */
      /* note: 512 temporary */
      if(log_buf[6] % 512) {
	/*
	 * not the first entry in this block
	 * we have the previous entry's size, set it
	 */
	short temp = log_buf[7];
	memcpy(dir_bp->b_un.b_buf + offset - log_buf[7] + 4,
	       &(temp), sizeof(short));
      } else {
	/*
	 * we are the first entry in the directory, update the size?
	 */
	/* note: 512 temporary */
	i = ((offset / 512) + 1) * 512;
	dir_p->di_size = (i > dir_p->di_size) ? i : dir_p->di_size;
      }

      /*
       * set the directory inode as dirty
       */
      inodirty();

      /* 
       * write out the block
       */
      dirty(dir_bp);
      dir_bp->b_flags &= ~B_INUSE;

      /*
       * update the associated inode
       */
      file_p = ginode(dir_entry.d_ino);
      file_p->di_nlink = log_buf[5];
      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[9];
    } else if(flag == BUF_UNDO) {
      /*
       * don't have to undo this operation, since it is a finishing
       * operation, rather than a starting or intermediary
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_REMDIRENTRY):
    if(flag == BUF_EXECUTE) {
      /*
       * clear the directory entry at the given offset
       */
      dir_p = ginode(log_buf[3]);
      dir_p->di_nlink = log_buf[4];
      inum = free_dir_entry(dir_p, log_buf[7], log_buf[8], log_buf[9]);
      inodirty();

      /*
       * set the link count of the associated inode
       */
      file_p = ginode(log_buf[5]);
      file_p->di_nlink = log_buf[6];
      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      if(log_buf[6] == 0) {
	return_val = LOGGING_FREEINODE;
      } else {
	return_val = BUF_ERROR;
      }
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[5];
    } else if(flag == BUF_UNDO) {
      /*
       * replace the directory entry somehow? 
       * (might have to store the entry in the log...)
       *

       * in this case, it should be possible, that if the inode was never
       * written causing the FREEINODE logging to occur that we can grab
       * all the related blocks from the inode (it was never zeroed!)
       * and then do everything correctly anyhow...
       */

      /*
       * do the normal stuff to remove an entry
       */
      dir_p = ginode(log_buf[3]);
      dir_p->di_nlink = log_buf[4];
      inum = free_dir_entry(dir_p, log_buf[7], log_buf[8], log_buf[9]);
      inodirty();

      /*
       * the inode must be destroyed, or else we wouldn't have waited to
       * do this... so no need to set the link count of the inode
       */

      /*
       * grab the inode, and read off all of the associated blocks
       */
      file_p = ginode(log_buf[5]);
      i = 0;
      while(j = get_block(file_p, i) /* get all the blocks */) {
	/* free the block */
	if((file_p->di_size - i * fs->fs_bsize) >= (uint)(fs->fs_bsize)) {
	  k = 8;
	} else {
	  k = ((file_p->di_size - i * fs->fs_bsize) / fs->fs_fsize) + 1;
	}
	free_frags(j, k);
	clear_block(file_p, i);
      }
      /* the last part might be a fragment... check the block count */
      
      /*
       * zero out the inode
       */
      memset(file_p, 0, sizeof(struct dinode));
      inodirty();

      /*
       * update the inode map to free the inode
       */
      free_inode(log_buf[5]);
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_CHANGEDIRENTRY):
    if(flag == BUF_EXECUTE) {
      /*
       * change the directory entry
       */
      dir_p = ginode(log_buf[3]);
      dir_p->di_nlink = log_buf[4];
      inum = change_dir_entry(dir_p, log_buf[7], log_buf[8]);

      /*
       * update the associated inode link counts
       */
      file_p = ginode(log_buf[5]);
      file_p->di_nlink = log_buf[6];
      inodirty();

      file_p = ginode(log_buf[8]);
      file_p->di_nlink = log_buf[9];
      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      if(log_buf[6] == 0) {
	return_val = LOGGING_FREEINODE;
      } else {
	return_val = BUF_ERROR;
      }
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[5];
    } else if(flag == BUF_UNDO) {
      /*
       * in this case, it should be possible, that if the inode was never
       * written causing the FREEINODE logging to occur that we can grab
       * all the related blocks from the inode (it was never zeroed!)
       * and then do everything correctly anyhow...
       */

      /*
       * note: we wouldn't have waited unless the link count of the losing
       * inode was 0, so we must delete it
       */

      /*
       * do the normal changedir stuff here
       */
      /*
       * change the directory entry
       */
      dir_p = ginode(log_buf[3]);
      dir_p->di_nlink = log_buf[4];
      inum = change_dir_entry(dir_p, log_buf[7], log_buf[8]);

      /*
       * update the associated inode link count (not the old one, just the new)
       */
      file_p = ginode(log_buf[8]);
      file_p->di_nlink = log_buf[9];
      inodirty();

      /*
       * delete the old inode
       */
      /*
       * grab the inode, and read off all of the associated blocks
       */
      file_p = ginode(log_buf[5]);
      i = 0;
      while(j = get_block(file_p, i) /* get all the blocks */) {
	/* free the block */
	if((file_p->di_size - i * fs->fs_bsize) >= (uint)(fs->fs_bsize)) {
	  k = 8;
	} else {
	  k = ((file_p->di_size - i * fs->fs_bsize) / fs->fs_fsize) + 1;
	}
	free_frags(j, k);
	clear_block(file_p, i);
      }
      /* the last part might be a fragment... check the block count */
      
      /*
       * zero out the inode
       */
      memset(file_p, 0, sizeof(struct dinode));
      inodirty();

      /*
       * update the inode map to free the inode
       */
      free_inode(log_buf[5]);
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_ADDINDIRPAGE):
    if(flag == BUF_EXECUTE) {
      /*
       * check if there is any data we should update the block with
       */
      get_block_list(l, log_buf[5]);

      /*
       * set the logical block number
       */
      file_p = ginode(log_buf[3]);
      set_block(file_p, log_buf[4], log_buf[5]);
      inodirty();
      
      /*
       * note: I think this is taken care of above now
       * WARNING: POTENTIALLY DANGEROUS TO ZERO THIS OUT
       * also shouldn't have to...
       */ /*
	    bp = getdatablk(log_buf[5], fs->fs_bsize);
	    memset(bp->b_un.b_buf, 0, fs->fs_bsize);
	    dirty(bp);
	    bp->b_flags &= ~B_INUSE;
	  */
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = LOGGING_CHANGESIZE;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * set the inode logical block number back to 0
       * (could it have ever been non-zero? I'm not sure...)
       * note: do anything at all? inode never written out...
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_ADDDIRECT):
    if(flag == BUF_EXECUTE) {
      /*
       * if the block will be an indirect pointer block, do some
       * extra work to ensure correctness
       */
      if(log_buf[4] >= NDADDR) {
	/* save the old data from the block */
	put_block_list(l, log_buf[5]);
      } else {
	get_block_list(l, log_buf[5]);
      }

      /*
       * set the direct block to the correct physical block
       */
      file_p = ginode(log_buf[3]);
      file_p->di_db[log_buf[4]] = log_buf[5];
      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = LOGGING_CHANGESIZE;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * set the direct block back to 0
       * (could it have been non-zero? I'm not sure...)
       * note: do anything at all? inode never written out...
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_UPDATEBLKMAP):
    if(flag == BUF_EXECUTE) {
      /*
       * updating the block map
       */
      alloc_frags(log_buf[4], log_buf[5]);
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = LOGGING_CHANGESIZE;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * set the given blocks to be free
       */
      free_frags(log_buf[4], log_buf[5]);
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_UPDATEINOMAP):
    if(flag == BUF_EXECUTE) {
      /*
       * update the inode map
       */
      alloc_inode(log_buf[3]);

      /*
       * initialize the inode
       */
      file_p = ginode(log_buf[3]);
      file_p->di_flags = 0;
      file_p->di_mode = log_buf[4];
      file_p->di_atime = file_p->di_ctime = file_p->di_mtime = log_buf[1];
      file_p->di_atimensec = file_p->di_ctimensec = 
	file_p->di_mtimensec = log_buf[2];
      file_p->di_size = 0;
      file_p->di_blocks = 0;
      for(i = 0; i < NDADDR + NIADDR; i++) {
	file_p->di_db[i] = 0;
      }
      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = LOGGING_ADDDIRENTRY;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * zero out the inode
       */
      file_p = ginode(log_buf[3]);
      memset(file_p, 0, sizeof(struct dinode));
      inodirty();

      /*
       * set the given inode to be free
       */
      free_inode(log_buf[3]);
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_MOVDIRENTRY):
    if(flag == BUF_EXECUTE) {
      /*
       * note: same as add dir entry, we differentiate because
       * this one can be done regardless of other concerns;
       * i.e. no need for undo, never does any ops before itself
       */

      /*
       * get the directory entry
       */
      memcpy(&dir_entry, buffer + sizeof(int)*9, 8);
      memcpy(&(dir_entry.d_name), buffer + sizeof(int)*9 + 8,
	     dir_entry.d_namlen);
      dir_entry.d_name[dir_entry.d_namlen] = 0;

      /* 
       * get the dinode, set the link count and the lbn
       */
      dir_p = ginode(log_buf[3]);
      dir_p->di_nlink = log_buf[4];
      dlbn = log_buf[6]/fs->fs_bsize;
      offset = log_buf[6] - dlbn * fs->fs_bsize;

      /*
       * update the directory entry
       */
      dir_bp = getdatablk(get_block(dir_p, dlbn), fs->fs_bsize);
      memcpy(dir_bp->b_un.b_buf + offset, &dir_entry, 8+dir_entry.d_namlen+1);

      /*
       * need to correct the previous entry's rec_len
       */
      /* note: 512 temporary */
      if(log_buf[6] % 512) {
	/*
	 * not the first entry in this block
	 * we have the previous entry's size, set it
	 */
	short temp = log_buf[7];
	memcpy(dir_bp->b_un.b_buf + offset - log_buf[7] + 4,
	       &(temp), sizeof(short));
      } else {
	/*
	 * we are the first entry in the directory, update the size?
	 */
	/* note: 512 temporary */
	dir_p->di_size = ((offset / 512) + 1) * 512;
      }

      /*
       * set the directory inode as dirty
       */
      inodirty();

      /* 
       * write out the block
       */
      dirty(dir_bp);
      dir_bp->b_flags &= ~B_INUSE;

      /*
       * update the associated inode
       */
      file_p = ginode(dir_entry.d_ino);
      file_p->di_nlink = log_buf[5];
      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      /* these can and should be executed immediately! */
      return_val = BUF_ERROR;
    } else if(flag == BUF_UNDO) {
      /*
       * don't have to undo this operation, since it is a finishing
       * operation, rather than a starting or intermediary
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_SETINDIRMETA):
    if(flag == BUF_EXECUTE) {
      /*
       * if the block will be an indirect pointer block, do some
       * extra work to ensure correctness by saveing the old data
       * from the block
       */
      put_block_list(l, log_buf[6]);

      /*
       * set the indirect meta data pointer block
       */
      bp = getdatablk(log_buf[5], fs->fs_bsize);
      memcpy(bp->b_un.b_data + log_buf[4], &log_buf[6], sizeof(int));

      dirty(bp);
      bp->b_flags &= ~B_INUSE;
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = LOGGING_CHANGESIZE;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * set the indirect pointer back to 0
       * (could it have been anything beside zero?  I don't think so...)
       * note: do anything at all? inode never written out...
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_FREEBLOCKS):
    if(flag == BUF_EXECUTE) {
      /*
       * clear all the inode block pointers
       */
      first_block = (log_buf[5] == 0) ? 0 : ((log_buf[5]-1)/fs->fs_bsize) + 1;
      last_block = (log_buf[4] / fs->fs_bsize);

      file_p = ginode(log_buf[3]);
      for(k = first_block; k <= last_block; k++) {
	/*
	 * clear the inode block pointer
	 */
	clear_block(file_p, k);
      }
      
      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      if(log_buf[5] == 0) {
	return_val = LOGGING_FREEINODE;
      } else {
	return_val = LOGGING_CHANGESIZE;
      }
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * I don't think we can undo this operation... is
       * that a problem?
       *
       * it should be the case that if the inode on disk is zero'ed out
       * then it was written, and either a CHANGESIZE or FREEINODE would
       * have occured, so we don't have to undo the operation, just read
       * off the inode and free the blocks appropriately
       */
      /*
       * however, the indirect pointer blocks might have been written
       * in which case, remaining operations need to be executed to allow
       * for proper remove during the CHANGEDIRENTRY or REMDIRENTRY
       * undo phases?
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_FREEFRAGS):
    if(flag == BUF_EXECUTE) {
      /*
       * free y fragments at block x
       */
      free_frags(log_buf[4], log_buf[5]);
    } else if(flag == BUF_GET_WAITFOR) {
      /* will ALWAYS occur after a LOGGING_FREEBLOCKS, so not important */
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * allocate those fragments back... although if we are never undoing
       * a FREEBLOCKS, we shouldn't undo this...
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_CHANGESIZE):
    if(flag == BUF_EXECUTE) {
      /*
       * set the size and the number of blocks in the inode
       */
      file_p = ginode(log_buf[3]);
      file_p->di_size = log_buf[4];
      file_p->di_blocks = log_buf[5];
      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * don't have to undo this operation, since it is a finishing
       * operation, rather than a starting or intermediary
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_SYMLINK):
    if(flag == BUF_EXECUTE) {
      /*
       * set the size of the inode
       */
      file_p = ginode(log_buf[3]);
      file_p->di_size = log_buf[4];

      /*
       * set the direct blocks to hold the given name
       */
      memcpy(file_p->di_db, buffer + sizeof(int)*6, log_buf[4]);

      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * don't have to undo this operation, since it is a finishing
       * operation, rather than a starting or intermediary
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_SETATTR):
    if(flag == BUF_EXECUTE) {
      /*
       * set all the attributes and the correct access/modify times
       */
      file_p = ginode(log_buf[3]);

      /* should I be keeping the actual times? probably... */
      file_p->di_atime = file_p->di_mtime = file_p->di_ctime = log_buf[1];
      file_p->di_atimensec = file_p->di_mtimensec = 
	file_p->di_ctimensec = log_buf[2];
      
      /* set the rest */
      file_p->di_mode = log_buf[4];
      file_p->di_uid = log_buf[5];
      file_p->di_gid = log_buf[6];
      file_p->di_flags = log_buf[7];

      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * don't have to undo this operation, since it is done after
       * the directory entry is created
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_MKNOD):
    if(flag == BUF_EXECUTE) {
      /*
       * set the rdev entry of the inode
       */
      file_p = ginode(log_buf[3]);
      file_p->di_rdev = log_buf[4];
      inodirty();
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {
      /*
       * don't have to undo this operation, since it is done after
       * the directory entry is created
       */
    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_INITDIR):
    if(flag == BUF_EXECUTE) {
      /*
       * set the two initial entries
       */
      dir_p = ginode(log_buf[5]);
      dir_p->di_nlink = log_buf[6];
      inodirty();

      file_p = ginode(log_buf[3]);
      file_p->di_nlink = log_buf[4];
      bp = getdatablk(get_block(file_p, 0), fs->fs_bsize);
      inodirty();

      dir_entry.d_ino = log_buf[3];
      dir_entry.d_reclen = 12;
      dir_entry.d_type = DT_DIR;
      dir_entry.d_namlen = 1;
      dir_entry.d_name[0] = '.';
      dir_entry.d_name[1] = 0;
      dir_entry.d_name[2] = 0;
      dir_entry.d_name[3] = 0;
      memcpy(bp->b_un.b_buf, &dir_entry, 12);

      dir_entry.d_ino = log_buf[5];
      dir_entry.d_reclen = 500;
      dir_entry.d_type = DT_DIR;
      dir_entry.d_namlen = 2;
      dir_entry.d_name[0] = '.';
      dir_entry.d_name[1] = '.';
      dir_entry.d_name[2] = 0;
      dir_entry.d_name[3] = 0;
      memcpy(bp->b_un.b_buf+12, &dir_entry, 12);

      dirty(bp);
      bp->b_flags &= ~B_INUSE;
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {

    } else {
      /* some sort of error... */
    }
    break;
  case(LOGGING_REALLOCBUF):
    if(flag == BUF_EXECUTE) {
      /*
       * copy the data from one block to another
       */
      bp = getdatablk(log_buf[5], fs->fs_bsize);
      memcpy(bp->b_un.b_buf, buffer + sizeof(int) * 9, log_buf[6]);

      /* do this?
	 memset(bp->b_un.b_buf + log_buf[6], 0, log_buf[7] - log_buf[6]);
      */
      
      dirty(bp);
      bp->b_flags &= ~B_INUSE;
    } else if(flag == BUF_GET_WAITFOR) {
      return_val = BUF_ERROR;
    } else if(flag == BUF_GET_DEPENDENT_VAL) {
      return_val = log_buf[3];
    } else if(flag == BUF_UNDO) {

    } else {
      /* some sort of error... */
    }
    break;
  default:
    /* this indicates a bad entry, return ERROR */
    break;
  }

  return return_val;
}
